{-# LANGUAGE CPP, ForeignFunctionInterface, FlexibleInstances#-}
module BlueZDefs 
    where

import Foreign
import Foreign.C.String
import Foreign.C.Types
import System.IO

import Data.Word

#include "bluetooth.h"
#include "hci.h"
#include "hci_lib.h"
#include <sys/socket.h>
#include "rfcomm.h"
#include "l2cap.h"

data InquiryInfo = InquiryInfo {bdaddr :: BDAddress,
                                pscan_rep_mode :: Int,
                                pscan_period_mode :: Int,
                                pscan_mode :: Int,
                                dev_class :: Ptr CUChar,
                                clock_offset :: Int}
                 deriving (Show,Eq)

termII :: InquiryInfo
termII = InquiryInfo nullPtr 0 0 0 nullPtr 0

instance Storable InquiryInfo where
    sizeOf _ = #const INQUIRY_INFO_SIZE
    alignment _ = alignment (undefined :: CUInt)
    peek ptr = do
      b <- (#peek inquiry_info, bdaddr) ptr
      pscan_rep <- fmap fromIntegral $ ((#peek inquiry_info, pscan_rep_mode) ptr :: IO CUChar)
      pscan_period <- fmap fromIntegral $ ((#peek inquiry_info, pscan_period_mode) ptr :: IO CUChar)
      pscan_m <- fmap fromIntegral $ ((#peek inquiry_info, pscan_mode) ptr :: IO CUChar)
      dev <- (#peek inquiry_info, dev_class) ptr
      clock <- fmap fromIntegral $ ((#peek inquiry_info, clock_offset) ptr :: IO CUInt)
      return $ InquiryInfo b pscan_rep pscan_period pscan_m dev clock
    poke ptr (InquiryInfo b prep pperiod pmode dev clock) = do
                                (#poke inquiry_info, bdaddr) ptr b
                                (#poke inquiry_info, pscan_rep_mode) ptr (fromIntegral prep :: CUChar)
                                (#poke inquiry_info, pscan_period_mode) ptr (fromIntegral pperiod :: CUChar)
                                (#poke inquiry_info, pscan_mode) ptr (fromIntegral pmode :: CUChar)
                                (#poke inquiry_info, dev_class) ptr dev
                                (#poke inquiry_info, clock_offset) ptr (fromIntegral clock :: CUInt)

-- Not sure if this should be just a type or a newtype with some extra instances built upon it.
-- keep that distinction in mind if you end up not using the ba2str & str2ba functions
type BDAddress = Ptr ()

type Address = [Word8]

bdAddrAny :: Address
bdAddrAny = [0,0,0,0,0,0]

afBluetooth :: CUShort
afBluetooth = #const AF_BLUETOOTH

rfcommProto :: CInt
rfcommProto = #const BTPROTO_RFCOMM

l2capProto :: CInt
l2capProto = #const BTPROTO_L2CAP

sockStream :: CInt
sockStream = #const SOCK_STREAM

sockSeqpacket :: CInt
sockSeqpacket = #const SOCK_SEQPACKET

-- newtype'd for the same damn reasons it's done in Network
-- namely because you don't want to accidentally allow 
-- people to get away with not converting it into proper
-- ordering
newtype BluePort = BluePort Word16

data RFComm = RFComm
data L2Cap = L2Cap

newtype BlueSock a = BlueSock CInt

data BlueAddr a = BlueAddr BluePort Address

instance Storable (BlueAddr L2Cap) where
    sizeOf _ = (#size struct sockaddr_l2)
    alignment _ = alignment (undefined :: CUShort)
    peek p = do
      port <- (#peek struct sockaddr_l2, l2_psm) p
      addr <- blueToAddress (2*sizeOf (undefined :: CUShort)) p
      return $ BlueAddr (BluePort port) addr
    poke p (BlueAddr (BluePort port) addr) = do
       (#poke struct sockaddr_l2, l2_family) p afBluetooth
       (#poke struct sockaddr_l2, l2_psm) p (fromIntegral port :: CUShort)
       let offset = 2*sizeOf (undefined :: CUShort)
       addressToBD offset p addr
      

instance Storable (BlueAddr RFComm) where
    sizeOf _ = (#size struct sockaddr_rc)
    alignment _ = alignment (undefined :: CUShort)
    peek p = do
        port <- (#peek struct sockaddr_rc, rc_channel) p
        addr <- blueToAddress (sizeOf (undefined :: CUShort)) p
        return $ BlueAddr (BluePort port) addr
    poke p (BlueAddr (BluePort port) addr) = do
       (#poke struct sockaddr_rc, rc_family) p afBluetooth
       (#poke struct sockaddr_rc, rc_channel) p (fromIntegral port :: CUInt)
       let offset = sizeOf (undefined :: CUShort)
       addressToBD offset p addr

bdToAddress :: BDAddress -> IO Address
bdToAddress = peekArray 6 . castPtr

addressToBD :: Int -> Ptr a -> Address -> IO ()
addressToBD offset p [] = return ()
addressToBD offset p (x:xs) = do
                       pokeByteOff p offset x
                       addressToBD (offset+1) p xs
blueToAddress :: Int -> Ptr a -> IO Address
blueToAddress offset p = mapM (\i -> peekByteOff p (offset+i)) [0..5]



